#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/tty.h>
#include <linux/ioport.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/cdev.h>

MODULE_LICENSE("GPL");

#define DEVNAME "HD44780_LCD"
static dev_t  dev;
static struct cdev my_cdev;
static int    DevMajor;
static int    isOpened = 0;

/* LOOKING FOR COMMAND LINE ARGUMENTS */
static int   LPT_PORT = 0x378;  // DEFAULT PORT IS LPT1
module_param(LPT_PORT, int, 0); // CAN OVERRIDDEN BY COMMAND LINE ARGUMENT (LPT_PORT=xxxx)


/* INCLUDE THE DRIVER CORE */
#include "hd44780.h"

/* CONTAINS THE POINTER TO OUR TTY CONSOLE (MAYBE XTERM) */
static char    msg[200];
int print_msg(char*);

/* FUNCTION PROTOTYPES FOR DEVICE OPERATIONS */
static int     device_open(struct inode*, struct file*);
static int     device_release(struct inode*, struct file*);
static ssize_t device_write(struct file*, const char*, size_t, loff_t*);


/* CREATING A FILE OPERATION STRUCTURE TO REGISTER OUR CHARACTER DEVICE */
static struct file_operations fops = {
    .write   = device_write,
    .open    = device_open,
    .release = device_release,
    .owner   = THIS_MODULE
};


static int device_open(struct inode* in, struct file* f)
{
    if (isOpened) return -EBUSY;
    print_msg("open()\n\r");
    isOpened = 1;
    return 0;
}

static int device_release(struct inode* in, struct file* f)
{
    print_msg("release()\n\r");
    isOpened = 0;;
    return 0;
}

static ssize_t device_write(struct file* f, const char* buf, size_t len, loff_t* off)
{
    int i;
    
    print_msg("write()\n\r");
    
    for (i = 0; i < len; ++i)
    {
	switch(buf[i])
	{
	case 9:
	    lcd_clear();
	    break;
	case 10:
	case 13:
	    lcd_gotoxy(0,1);
	    break;
	
	default:
	    lcd_putc(buf[i]);
	}
    }
    
    return len;
}


int print_msg(char* str)
{
    struct tty_struct *driver_tty = current->signal->tty;
    if (!driver_tty) return -ENOTTY;
    (driver_tty->driver->write)(driver_tty, str, strlen(str));
    return 0;
}


/* MODULE INITIALISATION */
int LCD_driver_init(void)
{
    int res;
    
    res = alloc_chrdev_region(&dev, 0, 1, DEVNAME);
    if (res < 0) return res;
    
    DevMajor = MAJOR(dev);
    
    cdev_init(&my_cdev, &fops);
    my_cdev.owner = THIS_MODULE;
    
    res = cdev_add(&my_cdev, dev, 1);
    if (res < 0)
    {
	unregister_chrdev_region(dev, 1);
	return res;
    }

    // INIT LCD
    
    lcd_init();
    lcd_write("HD44780 Driver", 14);
    lcd_gotoxy(0,1);
    lcd_write("Loaded! :-)", 11);
    
    print_msg("HD44780 LCD Driver loaded.\n\r");
    sprintf(msg, "Major number: %d.\n\r", DevMajor);
    print_msg(msg);

    return 0;
}

/* REMOVE MODULE */
void LCD_driver_exit(void)
{
    lcd_clear();
    
    cdev_del(&my_cdev);
    unregister_chrdev_region(dev, 1);

    sprintf(msg, "HD44780 LCD Driver unloaded.\n\r");

    print_msg(msg);
}

module_init(LCD_driver_init);
module_exit(LCD_driver_exit);

